from .eagle import EagleSpeculator, EagleSpeculatorConfig
from .independent import IndependentSpeculatorConfig
from .mlp import MLPSpeculatorConfig

__all__ = [
    "EagleSpeculator",
    "EagleSpeculatorConfig",
    "IndependentSpeculatorConfig",
    "MLPSpeculatorConfig",
]
